@@ -17,11 +17,8 @@
  *   GNU General Public License for more details.
  *
  *   You should have received a copy of the GNU General Public License
- *   along with this program; if not, write to the Free Software
- *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
- *
- *   msmtp is released under the GPL with the additional exemption that
- *   compiling, linking, and/or using OpenSSL is allowed.
+ *   along with this program; if not, write to the Free Software Foundation,
+ *   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -33,6 +30,7 @@
 #include <limits.h>
 #include <stdarg.h>
 #include <string.h>
+#include <strings.h>
 #include <ctype.h>
 #include <errno.h>
 extern int errno;
@@ -44,13 +42,16 @@ extern int errno;
 #include "crypto.h"
 #endif
 
+#include "gettext.h"
 #include "xalloc.h"
+#include "xvasprintf.h"
 
 #include "list.h"
 #include "net.h"
 #ifdef HAVE_SSL
 #include "tls.h"
 #endif /* HAVE_SSL */
+#include "stream.h"
 #include "smtp.h"
 
 
@@ -72,18 +73,26 @@ extern int errno;
 
 /* This defines the length of the *output* buffer for SMTP commands, without
  * CRLF. According to RFC 2821, SMTP commands may contain at most 512
- * characters including CRLF, thus the *maximum* size should be 510 characters.
- * Bus this is not sufficient for some authentication mechanisms, notably
+ * characters including CRLF, thus the maximum size should be 510 characters.
+ * But this is not sufficient for some authentication mechanisms, notably
  * GSSAPI, so this value must be higher.
  */
 #define SMTP_MAXCMDLEN 1022
 
+/* The maximum length of the SMTP command pipeline (the maximum number of
+ * commands that are sent at once, before starting to read the replies). This
+ * number should be large enough to keep the benefits of pipelining (saved round
+ * trips), and small enough to avoid problems (exceeding the TCP window size).
+ * A value of 1 disables pipelining.
+ */
+#define SMTP_PIPELINE_LIMIT 100
+
 /* This is the buffer length for copying the mail to the SMTP server.
  * According to RFC 2822, a line in a mail can contain at most 998 
  * characters + CRLF. Plus one character for '\0' = 1001 characters.
  * All lines should fit in a buffer of this size.
  * However, this length is not a limit; smtp_send_mail() will accept 
- * arbitrary long lines. 
+ * arbitrary long lines.
  */
 #define MAIL_BUFSIZE 1024
 
@@ -99,6 +108,7 @@ smtp_server_t smtp_new(FILE *debug, int protocol)
     smtp_server_t srv;
 
     srv.fd = -1;
+    net_readbuf_init(&(srv.readbuf));
 #ifdef HAVE_SSL
     tls_clear(&srv.tls);
 #endif /* HAVE_SSL */
@@ -117,9 +127,11 @@ smtp_server_t smtp_new(FILE *debug, int protocol)
  */
 
 int smtp_connect(smtp_server_t *srv, const char *host, int port, int timeout,
-	char *errstr)
+	char **server_canonical_name, char **server_address, 
+	char **errstr)
 {
-    return net_open_socket(host, port, timeout, &srv->fd, errstr);
+    return net_open_socket(host, port, timeout, &srv->fd, 
+	    server_canonical_name, server_address, errstr);
 }
 
 
@@ -137,7 +149,7 @@ int smtp_connect(smtp_server_t *srv, const char *host, int port, int timeout,
  * will be returned: SMTP_EIO, SMTP_EPROTO
  */
 
-int smtp_get_msg(smtp_server_t *srv, list_t **msg, char *errstr)
+int smtp_get_msg(smtp_server_t *srv, list_t **msg, char **errstr)
 {
     list_t *l;
     list_t *lp;
@@ -155,7 +167,8 @@ int smtp_get_msg(smtp_server_t *srv, list_t **msg, char *errstr)
 #ifdef HAVE_SSL
 	if (tls_is_active(&srv->tls))
 	{
-	    if (tls_gets(&srv->tls, line, SMTP_BUFSIZE, errstr) != TLS_EOK)
+	    if (tls_gets(&srv->tls, line, SMTP_BUFSIZE, &len, errstr) 
+		    != TLS_EOK)
 	    {
 		list_xfree(l, free);
 		return SMTP_EIO;
@@ -164,7 +177,8 @@ int smtp_get_msg(smtp_server_t *srv, list_t **msg, char *errstr)
 	else
 	{
 #endif /* HAVE_SSL */
-	    if (net_gets(srv->fd, line, SMTP_BUFSIZE, errstr) != NET_EOK)
+	    if (net_gets(srv->fd, &(srv->readbuf), line, SMTP_BUFSIZE, &len, 
+			errstr) != NET_EOK)
 	    {
 		list_xfree(l, free);
 		return SMTP_EIO;
@@ -172,7 +186,6 @@ int smtp_get_msg(smtp_server_t *srv, list_t **msg, char *errstr)
 #ifdef HAVE_SSL
 	}
 #endif /* HAVE_SSL */
-	len = strlen(line);
 	if (len < 4 
 		|| !(isdigit((unsigned char)line[0]) 
 		    && isdigit((unsigned char)line[1]) 
@@ -185,32 +198,38 @@ int smtp_get_msg(smtp_server_t *srv, list_t **msg, char *errstr)
 	     * message), but this is the term used in the RFCs.
 	     * An empty reply is a special case of an invalid reply - this
 	     * differentiation may help the user. */
-	    snprintf(errstr, errstr_bufsize, "the server sent an %s reply",
-		    (counter == 0 && len == 0) ? "empty" : "invalid");
+	    if (counter == 0 && len == 0)
+	    {
+		*errstr = xasprintf(_("the server sent an empty reply"));
+	    }
+	    else
+	    {
+		*errstr = xasprintf(_("the server sent an invalid reply"));
+	    }
 	    return SMTP_EPROTO;
 	}
-	/* kill CRLF */
-	line[len - 1] = '\0';
-	if (line[len - 2] == '\r')
+	if (srv->debug)
 	{
-	    line[len - 2] = '\0';
+	    fputs("<-- ", srv->debug);
+	    fwrite(line, sizeof(char), len, srv->debug);
+	}
+	/* kill CRLF */
+	line[--len] = '\0';
+	if (line[len - 1] == '\r')
+	{
+	    line[--len] = '\0';
 	}
 	list_insert(lp, xstrdup(line));
 	counter++;
 	lp = lp->next;
-	if (srv->debug)
-	{
-	    fprintf(srv->debug, "<-- %s\n", line);
-	}
     }
     while (line[3] == '-' && counter <= SMTP_MAXLINES);
 
     if (counter > SMTP_MAXLINES)
     {
 	list_xfree(l, free);
-	snprintf(errstr, errstr_bufsize, 
-		"Rejecting server reply that is longer than %d lines. "
-		"Increase SMTP_MAXLINES.", SMTP_MAXLINES);
+	*errstr = xasprintf(_("Rejecting server reply that is longer "
+		    "than %d lines. Increase SMTP_MAXLINES."), SMTP_MAXLINES);
 	return SMTP_EPROTO;
     }
 
@@ -235,30 +254,23 @@ int smtp_msg_status(list_t *msg)
 /*
  * smtp_put()
  *
- * This function writes a string to the SMTP server 'srv'. If 'newline' is set,
- * it will write TCP CRLF ('\r\n') after the string.
+ * This function writes 'len' characters from 's' to the SMTP server 'srv'.
  * Used error codes: SMTP_EIO
  */
 
-int smtp_put(smtp_server_t *srv, const char *s, int newline, char *errstr)
+int smtp_put(smtp_server_t *srv, const char *s, size_t len, char **errstr)
 {
     int e = 0;
 
 #ifdef HAVE_SSL
     if (tls_is_active(&srv->tls))
     {
-	if (tls_puts(&srv->tls, s, errstr) == TLS_EOK && newline)
-	{
-	    e = (tls_puts(&srv->tls, "\r\n", errstr) != TLS_EOK);
-	}
+	e = (tls_puts(&srv->tls, s, len, errstr) != TLS_EOK);
     }
     else
     {
 #endif /* HAVE_SSL */
-	if (net_puts(srv->fd, s, errstr) == NET_EOK && newline)
-	{
-	    e = (net_puts(srv->fd, "\r\n", errstr) != NET_EOK);
-	}
+	e = (net_puts(srv->fd, s, len, errstr) != NET_EOK);
 #ifdef HAVE_SSL
     }
 #endif /* HAVE_SSL */
@@ -268,7 +280,8 @@ int smtp_put(smtp_server_t *srv, const char *s, int newline, char *errstr)
     }
     if (srv->debug)
     {
-	fprintf(srv->debug, "--> %s\n", s);
+	fputs("--> ", srv->debug);
+	fwrite(s, sizeof(char), len, srv->debug);
     }
     
     return SMTP_EOK;
@@ -285,13 +298,15 @@ int smtp_put(smtp_server_t *srv, const char *s, int newline, char *errstr)
  * Used error codes: SMTP_EIO, SMTP_EINVAL
  */
 
-/* make gcc print format warnings for this function:
-int smtp_send_cmd(smtp_server_t *srv, char *errstr, const char *format, ...) 
+/* make gcc print format warnings for this function */
+#ifdef __GNUC__
+int smtp_send_cmd(smtp_server_t *srv, char **errstr, const char *format, ...) 
     __attribute__ ((format (printf, 3, 4)));
-*/
-int smtp_send_cmd(smtp_server_t *srv, char *errstr, const char *format, ...)
+#endif
+
+int smtp_send_cmd(smtp_server_t *srv, char **errstr, const char *format, ...)
 {
-    char line[SMTP_MAXCMDLEN + 1];
+    char line[SMTP_MAXCMDLEN + 3];
     int count;
     va_list args;
 
@@ -300,13 +315,15 @@ int smtp_send_cmd(smtp_server_t *srv, char *errstr, const char *format, ...)
     va_end(args);
     if (count >= SMTP_MAXCMDLEN + 1)
     {
-	snprintf(errstr, errstr_bufsize, 
-		"Cannot send command because it is "
-		"longer than %d characters. Increase SMTP_MAXCMDLEN.", 
+	*errstr = xasprintf(_("Cannot send command because it is "
+		"longer than %d characters. Increase SMTP_MAXCMDLEN."), 
 		SMTP_MAXCMDLEN);
 	return SMTP_EINVAL;
     }
-    return smtp_put(srv, line, 1, errstr);
+    line[count++] = '\r';
+    line[count++] = '\n';
+    line[count] = '\0';
+    return smtp_put(srv, line, (size_t)count, errstr);
 }
 
 
@@ -317,7 +334,7 @@ int smtp_send_cmd(smtp_server_t *srv, char *errstr, const char *format, ...)
  */
 
 int smtp_get_greeting(smtp_server_t *srv, list_t **errmsg, char **buf, 
-	char *errstr)
+	char **errstr)
 {
     int e;
     list_t *msg;
@@ -330,8 +347,7 @@ int smtp_get_greeting(smtp_server_t *srv, list_t **errmsg, char **buf,
     if (smtp_msg_status(msg) != 220)
     {
 	*errmsg = msg;
-	snprintf(errstr, errstr_bufsize, 
-		"Cannot get initial OK message from server");
+	*errstr = xasprintf(_("cannot get initial OK message from server"));
 	return SMTP_EPROTO;
     }
     if (buf)
@@ -353,7 +369,7 @@ int smtp_get_greeting(smtp_server_t *srv, list_t **errmsg, char **buf,
  */
 
 int smtp_init(smtp_server_t *srv, const char *ehlo_domain, list_t **errmsg, 
-	char *errstr)
+	char **errstr)
 {
     int e;
     list_t *ehlo_response;
@@ -368,27 +384,43 @@ int smtp_init(smtp_server_t *srv, const char *ehlo_domain, list_t **errmsg,
     *errmsg = NULL;
     if (srv->protocol == SMTP_PROTO_SMTP)
     {
-	e = smtp_send_cmd(srv, errstr, "EHLO %s", ehlo_domain);
+	if ((e = smtp_send_cmd(srv, errstr, "EHLO %s", ehlo_domain)) 
+		!= SMTP_EOK)
+	{
+	    return e;
+	}
+	if ((e = smtp_get_msg(srv, &ehlo_response, errstr)) != SMTP_EOK)
+	{
+	    return e;
+	}
+	if (smtp_msg_status(ehlo_response) != 250)
+	{
+	    /* fall back to HELO, for very old SMTP servers */
+	    list_xfree(ehlo_response, free);
+	    if ((e = smtp_send_cmd(srv, errstr, "HELO %s", ehlo_domain)) 
+		    != SMTP_EOK)
+	    {
+		return e;
+	    }
+	    if ((e = smtp_get_msg(srv, &ehlo_response, errstr)) != SMTP_EOK)
+	    {
+		return e;
+	    }
+	    if (smtp_msg_status(ehlo_response) != 250)
+	    {
+		*errmsg = ehlo_response;
+		*errstr = xasprintf(_("SMTP server does not accept "
+			    "EHLO or HELO commands"));
+		return SMTP_EPROTO;
+	    }
+	    list_xfree(ehlo_response, free);
+	    /* srv->cap.flags is 0 */
+	    return SMTP_EOK;
+	}
     }
-    else
+    else /* protocol is LMTP */
     {
-	e = smtp_send_cmd(srv, errstr, "LHLO %s", ehlo_domain);
-    }
-    if (e != SMTP_EOK)
-    {
-	return e;
-    }
-    if ((e = smtp_get_msg(srv, &ehlo_response, errstr)) != SMTP_EOK)
-    {
-	return e;
-    }
-    if (smtp_msg_status(ehlo_response) != 250)
-    {
-	/* This should only happen with very old SMTP servers that don't support
-	 * EHLO. It should never happen with LMTP. */
-	/* fall back to HELO */
-	list_xfree(ehlo_response, free);
-	if ((e = smtp_send_cmd(srv, errstr, "HELO %s", ehlo_domain)) 
+	if ((e = smtp_send_cmd(srv, errstr, "LHLO %s", ehlo_domain)) 
 		!= SMTP_EOK)
 	{
 	    return e;
@@ -400,13 +432,9 @@ int smtp_init(smtp_server_t *srv, const char *ehlo_domain, list_t **errmsg,
 	if (smtp_msg_status(ehlo_response) != 250)
 	{
 	    *errmsg = ehlo_response;
-	    snprintf(errstr, errstr_bufsize,
-		    "SMTP server does not accept EHLO or HELO commands");
+	    *errstr = xasprintf(_("command %s failed"), "LHLO");
 	    return SMTP_EPROTO;
 	}
-	list_xfree(ehlo_response, free);
-	/* srv->cap.flags is 0 */
-	return SMTP_EOK;
     }
 
     lp = ehlo_response;
@@ -440,8 +468,8 @@ int smtp_init(smtp_server_t *srv, const char *ehlo_domain, list_t **errmsg,
 	     * tell us about a maximum message size. Treat that as if the SIZE 
 	     * keyword was not seen. Also treat invalid numbers the same way.
 	     * The value 0 means there is no maximum.
-	     * See RFC 1653.
-	     */
+	     * See RFC 1653. */
+	    errno = 0;
 	    srv->cap.size = strtol(s + 8, &p, 10);
 	    if (!(*(s + 8) == '\0' || *p != '\0' || srv->cap.size < 0
 		    || (srv->cap.size == LONG_MAX && errno == ERANGE)))
@@ -503,7 +531,7 @@ int smtp_init(smtp_server_t *srv, const char *ehlo_domain, list_t **errmsg,
 
 #ifdef HAVE_SSL
 int smtp_tls_init(smtp_server_t *srv, const char *tls_key_file, 
-	const char *tls_ca_file, const char *tls_trust_file, char *errstr)
+	const char *tls_ca_file, const char *tls_trust_file, char **errstr)
 {
     return tls_init(&srv->tls, tls_key_file, tls_ca_file, tls_trust_file, 
 	    errstr);
@@ -518,7 +546,7 @@ int smtp_tls_init(smtp_server_t *srv, const char *tls_key_file,
  */
 
 #ifdef HAVE_SSL
-int smtp_tls_starttls(smtp_server_t *srv, list_t **error_msg, char *errstr)
+int smtp_tls_starttls(smtp_server_t *srv, list_t **error_msg, char **errstr)
 {
     int e;
     list_t *msg;
@@ -535,8 +563,7 @@ int smtp_tls_starttls(smtp_server_t *srv, list_t **error_msg, char *errstr)
     if (smtp_msg_status(msg) != 220)
     {
 	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize, 
-		"the server does not accept the STARTTLS command");
+	*errstr = xasprintf(_("command %s failed"), "STARTTLS");
 	return SMTP_EPROTO;
     }
     list_xfree(msg, free);
@@ -553,7 +580,7 @@ int smtp_tls_starttls(smtp_server_t *srv, list_t **error_msg, char *errstr)
 
 #ifdef HAVE_SSL
 int smtp_tls(smtp_server_t *srv, const char *hostname, int tls_nocertcheck,
-	tls_cert_info_t *tci, char *errstr)
+	tls_cert_info_t *tci, char **errstr)
 {
     return tls_start(&srv->tls, srv->fd, hostname, tls_nocertcheck, tci, 
 	    errstr);
@@ -571,7 +598,7 @@ int smtp_tls(smtp_server_t *srv, const char *hostname, int tls_nocertcheck,
 
 #ifndef USE_GSASL
 int smtp_auth_plain(smtp_server_t *srv, const char *user, const char *password,
-	list_t **error_msg, char *errstr)
+	list_t **error_msg, char **errstr)
 {
     char *s;
     char *b64;
@@ -609,15 +636,13 @@ int smtp_auth_plain(smtp_server_t *srv, const char *user, const char *password,
 	*error_msg = msg;
 	if (status == 504)
 	{
-	    snprintf(errstr, errstr_bufsize, 
-		    "AUTH PLAIN not accepted by the server");
+	    *errstr = xasprintf(_("command %s failed"), "AUTH PLAIN");
 	    return SMTP_EPROTO;
 	}
 	else
 	{
-	    snprintf(errstr, errstr_bufsize,
-		    "the server does not accept username/password "
-		    "(AUTH PLAIN)");
+	    *errstr = xasprintf(_("authentication failed (method %s)"), 
+		    "PLAIN");
 	    return SMTP_EAUTHFAIL;
 	}
     }
@@ -638,7 +663,7 @@ int smtp_auth_plain(smtp_server_t *srv, const char *user, const char *password,
 
 #ifndef USE_GSASL
 int smtp_auth_login(smtp_server_t *srv, const char *user, const char *password,
-	list_t **error_msg, char *errstr)
+	list_t **error_msg, char **errstr)
 {
     int e;
     list_t *msg;
@@ -659,8 +684,7 @@ int smtp_auth_login(smtp_server_t *srv, const char *user, const char *password,
     if (smtp_msg_status(msg) != 334)
     {
 	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize, 
-		"AUTH LOGIN not accepted by the server");
+	*errstr = xasprintf(_("command %s failed"), "AUTH LOGIN");
 	return SMTP_EPROTO;
     }
     list_xfree(msg, free);
@@ -681,8 +705,7 @@ int smtp_auth_login(smtp_server_t *srv, const char *user, const char *password,
     if (smtp_msg_status(msg) != 334)
     {
 	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize,
-		"the server does not accept username (AUTH LOGIN)");
+	*errstr = xasprintf(_("authentication failed (method %s)"), "LOGIN");
 	return SMTP_EAUTHFAIL;
     }
     list_xfree(msg, free);
@@ -703,8 +726,7 @@ int smtp_auth_login(smtp_server_t *srv, const char *user, const char *password,
     if (smtp_msg_status(msg) != 235)
     {
 	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize, 
-		"the server does not accept username/password (AUTH LOGIN)");
+	*errstr = xasprintf(_("authentication failed (method %s)"), "LOGIN");
 	return SMTP_EAUTHFAIL;
     }
     list_xfree(msg, free);
@@ -725,7 +747,7 @@ int smtp_auth_login(smtp_server_t *srv, const char *user, const char *password,
 #ifndef USE_GSASL
 int smtp_auth_cram_md5(smtp_server_t *srv, const char *user, 
 	const char *password,
-	list_t **error_msg, char *errstr)
+	list_t **error_msg, char **errstr)
 {
     unsigned char digest[16];
     char hex[] = "0123456789abcdef";
@@ -752,8 +774,7 @@ int smtp_auth_cram_md5(smtp_server_t *srv, const char *user,
     if (smtp_msg_status(msg) != 334)
     {
 	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize, 
-		"AUTH CRAM-MD5 not accepted by the server");
+	*errstr = xasprintf(_("command %s failed"), "AUTH CRAM-MD5");
 	return SMTP_EPROTO;
     }
     /* we know the line is at least 4 characters long */
@@ -764,8 +785,8 @@ int smtp_auth_cram_md5(smtp_server_t *srv, const char *user,
     if (!base64_decode(challenge, challenge_len, b64, &len))
     {
 	list_xfree(msg, free);
-	snprintf(errstr, errstr_bufsize,
-		"AUTH CRAM-MD5: server sent invalid challenge");
+	*errstr = xasprintf(_("authentication method CRAM-MD5: "
+		    "server sent invalid challenge"));
 	return SMTP_EPROTO;
     }
     list_xfree(msg, free);
@@ -802,9 +823,7 @@ int smtp_auth_cram_md5(smtp_server_t *srv, const char *user,
     if (smtp_msg_status(msg) != 235)
     {
 	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize, 
-		"the server does not accept username/password "
-		"(AUTH CRAM-MD5)");
+	*errstr = xasprintf(_("authentication failed (method %s)"), "CRAM-MD5");
 	return SMTP_EAUTHFAIL;
     }
     list_xfree(msg, free);
@@ -826,7 +845,7 @@ int smtp_auth_cram_md5(smtp_server_t *srv, const char *user,
 
 #ifndef USE_GSASL
 int smtp_auth_external(smtp_server_t *srv, const char *user,
-	list_t **error_msg, char *errstr)
+	list_t **error_msg, char **errstr)
 {
     size_t u_len;
     size_t b64_len;
@@ -846,8 +865,7 @@ int smtp_auth_external(smtp_server_t *srv, const char *user,
     if (smtp_msg_status(msg) != 334)
     {
 	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize, 
-		"AUTH EXTERNAL not accepted by the server");
+	*errstr = xasprintf(_("command %s failed"), "AUTH EXTERNAL");
 	return SMTP_EPROTO;
     }
     list_xfree(msg, free);
@@ -868,9 +886,7 @@ int smtp_auth_external(smtp_server_t *srv, const char *user,
     if (smtp_msg_status(msg) != 235)
     {
 	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize,
-		"the server does not accept username "
-		"(AUTH EXTERNAL)");
+	*errstr = xasprintf(_("authentication failed (method %s)"), "EXTERNAL");
 	return SMTP_EAUTHFAIL;
     }
     list_xfree(msg, free);
@@ -928,17 +944,10 @@ int smtp_client_supports_authmech(const char *mech)
     
 #else /* not USE_GSASL */
     
-    if (strcmp(mech, "CRAM-MD5") == 0
+    return (strcmp(mech, "CRAM-MD5") == 0
 	    || strcmp(mech, "PLAIN") == 0
 	    || strcmp(mech, "EXTERNAL") == 0
-	    || strcmp(mech, "LOGIN") == 0)
-    {
-	return 1;
-    }
-    else
-    {
-	return 0;
-    }
+	    || strcmp(mech, "LOGIN") == 0);
     
 #endif /* not USE_GSASL */
 }
@@ -954,11 +963,15 @@ int smtp_auth(smtp_server_t *srv,
 	const char *hostname,
 	const char *user, 
 	const char *password,
+#ifdef USE_GSASL
 	const char *ntlmdomain,
+#else
+	const char *ntlmdomain UNUSED,
+#endif
 	const char *auth_mech,
 	char *(*password_callback)(const char *hostname, const char *user),
 	list_t **error_msg,
-	char *errstr)
+	char **errstr)
 {
 #ifdef USE_GSASL
     int e;
@@ -974,25 +987,24 @@ int smtp_auth(smtp_server_t *srv,
 
 
     *error_msg = NULL;
-    if (strcmp(auth_mech, "") != 0 && 
-	    !smtp_server_supports_authmech(srv, auth_mech))
+    if (strcmp(auth_mech, "") != 0 
+	    && !smtp_server_supports_authmech(srv, auth_mech))
     {
-	snprintf(errstr, errstr_bufsize, 
-		"the server does not support authentication method %s", 
+	*errstr = xasprintf(
+		_("the server does not support authentication method %s"), 
 		auth_mech);
 	return SMTP_EUNAVAIL;
     }
     if ((error_code = gsasl_init(&ctx)) != GSASL_OK)
     {
-	snprintf(errstr, errstr_bufsize, "GNU SASL: %s", 
-		gsasl_strerror(error_code));
+	*errstr = xasprintf(_("GNU SASL: %s"), gsasl_strerror(error_code));
 	return SMTP_ELIBFAILED;
     }
     if (strcmp(auth_mech, "") != 0 && !gsasl_client_support_p(ctx, auth_mech))
     {
 	gsasl_done(ctx);
-	snprintf(errstr, errstr_bufsize, 
-		"GNU SASL: authentication method %s not supported", 
+	*errstr = xasprintf(
+		_("GNU SASL: authentication method %s not supported"), 
 		auth_mech);
 	return SMTP_ELIBFAILED;
     }
@@ -1043,14 +1055,13 @@ int smtp_auth(smtp_server_t *srv,
 	if (!tls_is_active(&srv->tls))
 	{
 #endif /* HAVE_SSL */
-  	    snprintf(errstr, errstr_bufsize,
-   		    "cannot use a secure authentication method");
+	    *errstr = xasprintf(_("cannot use a secure authentication method"));
 #ifdef HAVE_SSL
 	}
 	else
 	{
-   	    snprintf(errstr, errstr_bufsize,
-       		    "cannot find a usable authentication method");
+	    *errstr = xasprintf(
+		    _("cannot find a usable authentication method"));
      	}
 #endif /* not HAVE_SSL */
 	return SMTP_EUNAVAIL;
@@ -1059,8 +1070,7 @@ int smtp_auth(smtp_server_t *srv,
     if ((error_code = gsasl_client_start(ctx, auth_mech, &sctx)) != GSASL_OK)
     {
 	gsasl_done(ctx);
-	snprintf(errstr, errstr_bufsize, 
-		"GNU SASL: %s", gsasl_strerror(error_code));
+	*errstr = xasprintf(_("GNU SASL: %s"), gsasl_strerror(error_code));
 	return SMTP_ELIBFAILED;
     }
 
@@ -1118,8 +1128,7 @@ int smtp_auth(smtp_server_t *srv,
 	{
 	    gsasl_finish(sctx);
 	    gsasl_done(ctx);
-	    snprintf(errstr, errstr_bufsize, 
-		    "GNU SASL: %s", gsasl_strerror(error_code));
+	    *errstr = xasprintf(_("GNU SASL: %s"), gsasl_strerror(error_code));
 	    return SMTP_ELIBFAILED;
 	}
 	if (!input)
@@ -1165,8 +1174,7 @@ int smtp_auth(smtp_server_t *srv,
 		gsasl_finish(sctx);
 		gsasl_done(ctx);
 		free(outbuf);
-		snprintf(errstr, errstr_bufsize, 
-			"authentication failed (GNU SASL, method %s)", 
+		*errstr = xasprintf(_("authentication failed (method %s)"), 
 			auth_mech);
 		return SMTP_EAUTHFAIL;
 	    }	    
@@ -1205,8 +1213,7 @@ int smtp_auth(smtp_server_t *srv,
 		gsasl_finish(sctx);
 		gsasl_done(ctx);
 		free(outbuf);
-		snprintf(errstr, errstr_bufsize, 
-			"authentication failed (GNU SASL, method %s)", 
+		*errstr = xasprintf(_("authentication failed (method %s)"), 
 			auth_mech);
 		return SMTP_EAUTHFAIL;
 	    }	    
@@ -1222,8 +1229,7 @@ int smtp_auth(smtp_server_t *srv,
     {
 	gsasl_finish(sctx);
 	gsasl_done(ctx);
-	snprintf(errstr, errstr_bufsize, 
-		"authentication failed: %s (GNU SASL, method %s)", 
+	*errstr = xasprintf(_("authentication failed: %s (method %s)"), 
 		gsasl_strerror(error_code), auth_mech);
 	return SMTP_EAUTHFAIL;
     }
@@ -1244,8 +1250,7 @@ int smtp_auth(smtp_server_t *srv,
 	if (smtp_msg_status(msg) != 235)
 	{
 	    *error_msg = msg;
-	    snprintf(errstr, errstr_bufsize, 
-	    	    "authentication failed (GNU SASL, method %s)", 
+	    *errstr = xasprintf(_("authentication failed (method %s)"),
 		    auth_mech);
 	    return SMTP_EAUTHFAIL;
 	}
@@ -1260,11 +1265,11 @@ int smtp_auth(smtp_server_t *srv,
     
     
     *error_msg = NULL;
-    if (strcmp(auth_mech, "") != 0 && 
-	    !smtp_server_supports_authmech(srv, auth_mech))
+    if (strcmp(auth_mech, "") != 0 
+	    && !smtp_server_supports_authmech(srv, auth_mech))
     {
-	snprintf(errstr, errstr_bufsize, 
-		"the server does not support authentication method %s", 
+	*errstr = xasprintf( 
+		_("the server does not support authentication method %s"), 
 		auth_mech);
 	return SMTP_EUNAVAIL;
     }
@@ -1295,14 +1300,13 @@ int smtp_auth(smtp_server_t *srv,
        	if (!tls_is_active(&srv->tls))
  	{
 #endif /* HAVE_SSL */
-  	    snprintf(errstr, errstr_bufsize,
-   		    "cannot use a secure authentication method");
+	    *errstr = xasprintf(_("cannot use a secure authentication method"));
 #ifdef HAVE_SSL
     	}
     	else
  	{
-      	    snprintf(errstr, errstr_bufsize,
-      		    "cannot find a usable authentication method");
+	    *errstr = xasprintf(
+		    _("cannot find a usable authentication method"));
 	}
 #endif /* not HAVE_SSL */
 	return SMTP_EUNAVAIL;
@@ -1313,8 +1317,8 @@ int smtp_auth(smtp_server_t *srv,
 	/* CRAMD-MD5, PLAIN, LOGIN all need a user name and a password */
 	if (!user)
 	{
-	    snprintf(errstr, errstr_bufsize,
-	    	    "authentication method %s needs a user name", auth_mech);
+	    *errstr = xasprintf(_("authentication method %s needs a user name"),
+		    auth_mech);
 	    return SMTP_EUNAVAIL;
 	}
 	if (!password)
@@ -1322,8 +1326,8 @@ int smtp_auth(smtp_server_t *srv,
 	    if (!password_callback 
 		    || !(callback_password = password_callback(hostname, user)))
 	    {
-		snprintf(errstr, errstr_bufsize,
-		    	"authentication method %s needs a password", 
+		*errstr = xasprintf(
+			_("authentication method %s needs a password"),
 			auth_mech);
 		return SMTP_EUNAVAIL;
 	    }
@@ -1349,8 +1353,8 @@ int smtp_auth(smtp_server_t *srv,
     }
     else
     {
-	snprintf(errstr, errstr_bufsize, 
-		"authentication method %s not supported", auth_mech);
+	*errstr = xasprintf(_("authentication method %s not supported"),
+		auth_mech);
 	e = SMTP_ELIBFAILED;
     }
     free(callback_password);
@@ -1372,149 +1376,137 @@ int smtp_send_envelope(smtp_server_t *srv,
 	const char *dsn_notify,
 	const char *dsn_return,
 	list_t **error_msg,
-	char *errstr)
+	char **errstr)
 {
     int e;
-    int e_tmp;
-    list_t *msg;
-    list_t *lp;
+    list_t *msg;    
+    int mailfrom_cmd_was_sent = 0;
+    int mailfrom_reply_was_rcvd = 0;
+    list_t *rcpt_send = recipients;
+    list_t *rcpt_recv = recipients;
+    int data_cmd_was_sent = 0;
+    int data_reply_was_rcvd = 0;
+    int pipeline_limit = 1;
+    int piped_commands = 0;
+
     
     *error_msg = NULL;
-    /* tell the server who wants to send mail */
-    if (dsn_return)
-    {
-	e = smtp_send_cmd(srv, errstr, "MAIL FROM:<%s> RET=%s", 
-		envelope_from, dsn_return);
-    }
-    else
-    {
-	e = smtp_send_cmd(srv, errstr, "MAIL FROM:<%s>", envelope_from);
-    }
-    if (e != SMTP_EOK)
-    {
-	return e;
-    }
-    if (!(srv->cap.flags & SMTP_CAP_PIPELINING))
-    {
-	if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
-	{
-	    return e;
-	}
-	if (smtp_msg_status(msg) != 250)
-	{
-	    *error_msg = msg;
-	    snprintf(errstr, errstr_bufsize, 
-		    "envelope from address %s not accepted by the server", 
-		    envelope_from);
-	    return SMTP_EINVAL;
-	}
-	list_xfree(msg, free);
-    }
-    /* tell the server who to send the mail to */
-    lp = recipients;
-    while (!list_is_empty(lp))
-    {
-	lp = lp->next;
-	if (dsn_notify)
-	{
-	    e = smtp_send_cmd(srv, errstr, "RCPT TO:<%s> NOTIFY=%s", 
-		    (char *)(lp->data), dsn_notify);
-	}
-	else
-	{
-	    e = smtp_send_cmd(srv, errstr, "RCPT TO:<%s>", (char *)(lp->data));
-	}
-	if (e != SMTP_EOK)
-	{
-	    return e;
-	}
-	if (!(srv->cap.flags & SMTP_CAP_PIPELINING))
-	{
-	    if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
-	    {
-		return e;
-	    }
-	    if (smtp_msg_status(msg) != 250)
-	    {
-		*error_msg = msg;
-		snprintf(errstr, errstr_bufsize, 
-			"recipient address %s not accepted by the server", 
-			(char *)(lp->data));
-		return SMTP_EINVAL;
-	    }
-	    list_xfree(msg, free);
-	}
-    }
-    /* get all pipelined replies if pipelining is used */
     if (srv->cap.flags & SMTP_CAP_PIPELINING)
     {
-	if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
+	pipeline_limit = SMTP_PIPELINE_LIMIT;
+    }
+        
+    /* Send the MAIL FROM, RCPT TO and DATA commands using pipelining. The 
+     * number of pipelined commands will never be greater than pipeline_limit
+     * to avoid problems with the TCP window size (exceeding it can lead to 
+     * deadlocks). pipeline_limit == 1 disables pipelining. */
+    while (!data_reply_was_rcvd)
+    {
+	while (!data_cmd_was_sent && piped_commands < pipeline_limit)
 	{
-	    return e;
-	}
-     	if (smtp_msg_status(msg) != 250)
-	{
-	    *error_msg = msg;
-	    snprintf(errstr, errstr_bufsize,
-		    "envelope from address %s not accepted by the server", 
-		    envelope_from);
-	    e = SMTP_EINVAL;
-	}
-	else
-	{
-	    list_xfree(msg, free);
-	}
-	lp = recipients;
-	while (!list_is_empty(lp))
-	{
-	    lp = lp->next;
-	    if ((e_tmp = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
+	    /* send */
+	    if (!mailfrom_cmd_was_sent)
 	    {
-		return e_tmp;
-	    }
-    	    if (smtp_msg_status(msg) != 250)
-	    {
-		if (e == SMTP_EOK)
+		if (dsn_return)
 		{
-		    *error_msg = msg;
-		    e = SMTP_EINVAL; 
-		    snprintf(errstr, errstr_bufsize, 
-			    "recipient address %s not accepted by the server",
-			    (char *)(lp->data));
+		    e = smtp_send_cmd(srv, errstr, "MAIL FROM:<%s> RET=%s", 
+			    envelope_from, dsn_return);
 		}
 		else
 		{
-		    list_xfree(msg, free);
+		    e = smtp_send_cmd(srv, errstr, "MAIL FROM:<%s>", 
+			    envelope_from);
 		}
+		if (e != SMTP_EOK)
+		{
+		    return e;
+		}		
+		mailfrom_cmd_was_sent = 1;
+	    }
+	    else if (!list_is_empty(rcpt_send))
+	    {
+		rcpt_send = rcpt_send->next;
+		if (dsn_notify)
+		{
+		    e = smtp_send_cmd(srv, errstr, "RCPT TO:<%s> NOTIFY=%s",
+			    (char *)(rcpt_send->data), dsn_notify);
+		}
+		else
+		{
+		    e = smtp_send_cmd(srv, errstr, "RCPT TO:<%s>", 
+			    (char *)(rcpt_send->data));
+		}
+		if (e != SMTP_EOK)
+		{
+		    return e;
+		}		
 	    }
 	    else
 	    {
-		list_xfree(msg, free);
+		if ((e = smtp_send_cmd(srv, errstr, "DATA")) != SMTP_EOK)
+		{
+		    return e;
+		}
+		data_cmd_was_sent = 1;
 	    }
+	    piped_commands++;
 	}
-	if (e != SMTP_EOK)
+	while (piped_commands > 0)
 	{
-	    return e;
+	    /* receive */
+	    if (!mailfrom_reply_was_rcvd)
+	    {
+		if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
+		{
+		    return e;
+		}
+		if (smtp_msg_status(msg) != 250)
+		{
+		    *error_msg = msg;
+		    *errstr = xasprintf(_("envelope from address %s not "
+				"accepted by the server"), envelope_from);
+		    return SMTP_EINVAL;
+		}
+		list_xfree(msg, free);
+		mailfrom_reply_was_rcvd = 1;
+	    }
+	    else if (!list_is_empty(rcpt_recv))
+	    {
+		rcpt_recv = rcpt_recv->next;
+		if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
+		{
+		    return e;
+		}
+		if (smtp_msg_status(msg) != 250)
+		{
+		    *error_msg = msg;
+		    *errstr = xasprintf(_("recipient address %s not "
+				"accepted by the server"),
+			    (char *)(rcpt_recv->data));
+		    return SMTP_EINVAL;
+		}
+		list_xfree(msg, free);		
+	    }
+	    else
+	    {
+		if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
+		{
+		    return e;
+		}
+		if (smtp_msg_status(msg) != 354)
+		{
+		    *error_msg = msg;
+		    *errstr = xasprintf(
+			    _("the server does not accept mail data"));
+		    return SMTP_EUNAVAIL;
+		}
+		list_xfree(msg, free);
+		data_reply_was_rcvd = 1;
+	    }
+	    piped_commands--;
 	}
     }
-    /* announce the mail */
-    if ((e = smtp_send_cmd(srv, errstr, "DATA")) != SMTP_EOK)
-    {
-	return e;
-    }
-    if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
-    {
-	return e;
-    }
-    if (smtp_msg_status(msg) != 354)
-    {
-	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize, 
-		"the server does not accept the mail");
-	return SMTP_EUNAVAIL;
-    }
-    list_xfree(msg, free);
-    
+
     return SMTP_EOK;
 }
 
@@ -1526,36 +1518,64 @@ int smtp_send_envelope(smtp_server_t *srv,
  */
 
 int smtp_send_mail(smtp_server_t *srv, FILE *mailf, int keep_bcc, 
-	long *mailsize, char *errstr)
+	long *mailsize, char **errstr)
 {
-    char buffer[MAIL_BUFSIZE];
+    char bigbuffer[MAIL_BUFSIZE + 3];	/* buffer + leading dot + ending CRLF */
+    char *buffer;
     size_t len;
+    char *send_buf;
+    size_t send_len;
     int in_header;
     int in_bcc;
     int line_starts;
     int line_continues;
     int e;
     
+    bigbuffer[0] = '.';
+    buffer = bigbuffer + 1;
     in_header = 1;
     in_bcc = 0;
     line_continues = 0;
     e = SMTP_EOK;
-    while (fgets(buffer, (int)sizeof(buffer), mailf) != NULL)
+    for (;;)
     {
+	if (stream_gets(mailf, buffer, MAIL_BUFSIZE, &len, errstr) 
+		!= STREAM_EOK)
+	{
+	    return SMTP_EIO;
+	}
+	if (len == 0)
+	{
+	    break;
+	}
 	line_starts = !line_continues;
-	len = strlen(buffer);
 	if (len > 0 && buffer[len - 1] == '\n')
 	{
+	    /* first case: we have a line end */
 	    buffer[--len] = '\0';
+	    if (len > 0 && buffer[len - 1] == '\r')
+	    {
+		buffer[--len] = '\0';
+	    }
 	    line_continues = 0;
 	}
 	else if (len == MAIL_BUFSIZE - 1)
 	{
+	    /* second case: the line continues */
+	    if (buffer[len - 1] == '\r')
+	    {
+		/* We have CRLF that is divided by the buffer boundary. Since CR
+		 * may not appear alone in a mail according to RFC2822, we
+		 * know that the next buffer will be "\n\0", so it's safe to
+		 * just delete the CR. */
+		buffer[--len] = '\0';
+	    }
 	    line_continues = 1;
 	}
 	else
 	{
-	    /* last line lacks a newline character */
+	    /* third case: this is the last line, and it lacks a newline 
+	     * character */
 	    line_continues = 0;
 	}
 	if (!keep_bcc)
@@ -1570,7 +1590,7 @@ int smtp_send_mail(smtp_server_t *srv, FILE *mailf, int keep_bcc,
 		{
 		    if (!in_bcc)
 		    {
-			if (strncmp(buffer, "Bcc: ", 5) == 0)
+			if (strncasecmp(buffer, "Bcc:", 4) == 0)
 			{
 			    in_bcc = 1;
 			    /* remove Bcc header by ignoring this line */
@@ -1602,28 +1622,30 @@ int smtp_send_mail(smtp_server_t *srv, FILE *mailf, int keep_bcc,
 		}
 	    }
 	}
+	send_buf = buffer;
+	send_len = len;
 	if (line_starts && buffer[0] == '.')
 	{
-	    e = smtp_put(srv, ".", 0, errstr);
-	    /* do not count the quoting dot in 'mailsize' */
+	    /* Quote the leading dot with another dot */
+	    send_buf = bigbuffer;
+	    send_len = len + 1;
 	}
-	if (e == SMTP_EOK)
+	if (!line_continues)
 	{
-    	    e = smtp_put(srv, buffer, !line_continues, errstr);
-	    /* count a newline as one character */
-	    *mailsize += (long)len + (line_continues ? 0 : 1);
+	    /* Append CRLF */
+	    buffer[len] = '\r';
+	    buffer[len + 1] = '\n';
+	    buffer[len + 2] = '\0';
+	    send_len += 2;
 	}
-    	if (e != SMTP_EOK)
+	/* Update mailsize. Do not count a quote dot. Count CRLF as one
+	 * character. */
+	*mailsize += (long)len + (line_continues ? 0 : 1);
+	if ((e = smtp_put(srv, send_buf, send_len, errstr)) != SMTP_EOK)
 	{
 	    return e;
 	}
     }
-    if (ferror(mailf))
-    {
-	snprintf(errstr, errstr_bufsize, 
-		"input error while reading the mail");
-	return SMTP_EIO;
-    }
 
     return SMTP_EOK;
 }
@@ -1635,7 +1657,7 @@ int smtp_send_mail(smtp_server_t *srv, FILE *mailf, int keep_bcc,
  * see smtp.h
  */
 
-int smtp_end_mail(smtp_server_t *srv, list_t **error_msg, char *errstr)
+int smtp_end_mail(smtp_server_t *srv, list_t **error_msg, char **errstr)
 {
     int e;
     list_t *msg;
@@ -1652,8 +1674,7 @@ int smtp_end_mail(smtp_server_t *srv, list_t **error_msg, char *errstr)
     if (smtp_msg_status(msg) != 250)
     {
 	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize, 
-		"the server did not accept the mail");
+	*errstr = xasprintf(_("the server did not accept the mail"));
 	return SMTP_EUNAVAIL;
     }
     list_xfree(msg, free);
@@ -1678,7 +1699,7 @@ int smtp_end_mail_lmtp(smtp_server_t *srv,
 	list_t *recipients, 
 	list_t **errstrs,
 	list_t **error_msgs,
-	char *errstr)
+	char **errstr)
 {
     int e;
     list_t *msg;
@@ -1716,9 +1737,7 @@ int smtp_end_mail_lmtp(smtp_server_t *srv,
 	if (smtp_msg_status(msg) != 250)
 	{
 	    all_recipients_accepted = 0;
-	    tmp = xmalloc(errstr_bufsize * sizeof(char));
-	    snprintf(tmp, errstr_bufsize, 
-		    "the server refuses to send the mail to %s",
+	    tmp = xasprintf(_("the server refuses to send the mail to %s"),
 		    (char *)(lp_recipients->data));
 	    list_insert(lp_errstrs, tmp);
 	    list_insert(lp_error_msgs, msg);
@@ -1756,7 +1775,7 @@ int smtp_end_mail_lmtp(smtp_server_t *srv,
  */
 
 int smtp_etrn(smtp_server_t *srv, const char *etrn_argument, 
-	list_t **error_msg, char *errstr)
+	list_t **error_msg, char **errstr)
 {
     int e;
     list_t *msg;
@@ -1781,23 +1800,22 @@ int smtp_etrn(smtp_server_t *srv, const char *etrn_argument,
 	case 458: /* Unable to queue messages for node <x> */
 	case 459: /* Node <x> not allowed: <reason> */
 	    *error_msg = msg;
-	    snprintf(errstr, errstr_bufsize,
-		    "the server is unable to fulfill the request");
+	    *errstr = xasprintf(
+		    _("the server is unable to fulfill the request"));
 	    return SMTP_EUNAVAIL;
 	    break;
 	    
 	case 500: /* Syntax Error */
 	case 501: /* 501 Syntax Error in Parameters */
 	    *error_msg = msg;
-	    snprintf(errstr, errstr_bufsize,
-		    "invalid argument for Remote Message Queue Starting");
+	    *errstr = xasprintf(
+		    _("invalid argument for Remote Message Queue Starting"));
 	    return SMTP_EINVAL;
 	    break;
 	    
 	default:
 	    *error_msg = msg;
-	    snprintf(errstr, errstr_bufsize,
-		    "the server sent an invalid reply to the ETRN command");
+	    *errstr = xasprintf(_("command %s failed"), "ETRN");
 	    return SMTP_EPROTO;
 	    break;
     }
@@ -1813,7 +1831,7 @@ int smtp_etrn(smtp_server_t *srv, const char *etrn_argument,
  * see smtp.h
  */
 
-int smtp_quit(smtp_server_t *srv, char *errstr)
+int smtp_quit(smtp_server_t *srv, char **errstr)
 {
     int e;
     list_t *msg;
