@@ -658,6 +658,8 @@ int tls_check_cert(tls_t *tls, const char *hostname, int verify, char **errstr)
     /* If 'verify' is true, this function uses the trusted CAs in the 
      * credentials structure. So you must have installed one or more CA 
      * certificates. */
+    gnutls_certificate_set_verify_flags(tls->cred, 
+	    GNUTLS_VERIFY_ALLOW_X509_V1_CA_CRT);
     if ((error_code = gnutls_certificate_verify_peers2(tls->session, 
 		    &status)) != 0)
     {
@@ -1414,7 +1416,7 @@ void tls_close(tls_t *tls)
     if (tls->is_active)
     {
 #ifdef HAVE_GNUTLS
-	gnutls_bye(tls->session, GNUTLS_SHUT_RDWR);
+	gnutls_bye(tls->session, GNUTLS_SHUT_WR);
 	gnutls_deinit(tls->session);
 	gnutls_certificate_free_credentials(tls->cred);
 #endif /* HAVE_GNUTLS */
